//----------------------------------------------------------------------------------
// File:   SoftShadows.cpp
// Author: Kirill Dmitriev
// Email:  sdkfeedback@nvidia.com
// 
// Copyright (c) 2007 NVIDIA Corporation. All rights reserved.
//
// TO  THE MAXIMUM  EXTENT PERMITTED  BY APPLICABLE  LAW, THIS SOFTWARE  IS PROVIDED
// *AS IS*  AND NVIDIA AND  ITS SUPPLIERS DISCLAIM  ALL WARRANTIES,  EITHER  EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED  TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL  NVIDIA OR ITS SUPPLIERS
// BE  LIABLE  FOR  ANY  SPECIAL,  INCIDENTAL,  INDIRECT,  OR  CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION,  DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE  USE OF OR INABILITY  TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
//
//----------------------------------------------------------------------------------

//
// demonstrates use of SSMap class.
//

#include <dxut.h>
#include <dxutgui.h>
#include <dxutsettingsdlg.h>
#include <sdkmisc.h>

#include <nvutcamera.h>
#include <nvutmesh.h>
#include "SoftShadowMap.h"

#define MAX_WCHAR_SIZE      260

static NVUTCamera g_Camera;
static NVUTCamera g_LCamera;
static CD3DSettingsDlg g_D3DSettingsDlg;
static CDXUTDialogResourceManager g_DialogResourceManager;
static CDXUTDialog g_HUD;
static NVUTMesh g_MeshScene;
static float g_fSceneBoxDiag;
static NVUTMesh g_MeshLight;
static ID3DX10Font* g_pFont10 = NULL;
static ID3DX10Sprite* g_pSprite10 = NULL;
static CDXUTDialog g_SampleUI;
static ID3D10Effect *g_pEffect = NULL;
static ID3D10InputLayout *g_pMaxLayout = NULL;
static D3DXVECTOR3 g_vLightDir;
static ID3D10RasterizerState *g_pRenderState = NULL;
static SSMap ssmap;
static bool g_bShowUI = true;

//--------------------------------------------------------------------------------------
// UI control IDs
//--------------------------------------------------------------------------------------
#define IDC_TOGGLEFULLSCREEN    1
#define IDC_TOGGLEREF           3
#define IDC_CHANGEDEVICE        4
#define IDC_LIGHT_SIZE_LABEL    5
#define IDC_LIGHT_SIZE          6
#define IDC_SHADOW_ALGORITHM    7
#define IDC_SHADOW_ALGORITHM_LABEL 8
#define IDC_BTEXTURED 9
#define IDC_LOADNEWMODEL        10

//--------------------------------------------------------------------------------------
// Forward declarations 
//--------------------------------------------------------------------------------------
bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext );
void CALLBACK OnFrameMove( double fTime, float fElapsedTime, void* pUserContext );
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing, void* pUserContext );
void CALLBACK OnKeyboard( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext );
void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext );

bool CALLBACK IsD3D10DeviceAcceptable( UINT Adapter, UINT Output, D3D10_DRIVER_TYPE DeviceType, DXGI_FORMAT BackBufferFormat, bool bWindowed, void* pUserContext );
HRESULT CALLBACK OnD3D10CreateDevice( ID3D10Device* pDev10, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );
HRESULT CALLBACK OnD3D10SwapChainResized( ID3D10Device* pDev10, IDXGISwapChain *pSwapChain, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );
void CALLBACK OnD3D10FrameRender( ID3D10Device* pDev10, double fTime, float fElapsedTime, void* pUserContext );
void CALLBACK OnD3D10SwapChainReleasing( void* pUserContext );
void CALLBACK OnD3D10DestroyDevice( void* pUserContext );

//--------------------------------------------------------------------------------------
// Load new model
//--------------------------------------------------------------------------------------
static void LoadNewModel(bool bNeedUI = false)
{
    //HRESULT hr;              
    static bool bVirgin = true;
    static WCHAR g_wcCleanFileName[MAX_WCHAR_SIZE];
    if (bVirgin)
    {
        bVirgin = false;
        memcpy(g_wcCleanFileName, L"bonfire_small.x", sizeof(L"bonfire_small.x"));
    }

    BOOL bResult = TRUE;
    if (bNeedUI)
    {
        OPENFILENAME info;
        ZeroMemory(&info, sizeof (OPENFILENAME));
        info.lStructSize = sizeof (OPENFILENAME);
        info.hwndOwner = DXUTGetHWND();
        info.lpstrFilter = L"X Files\0*.x\0\0";
        info.nFilterIndex = 1;               
        info.lpstrFile = &g_wcCleanFileName[0];
        info.nMaxFile = MAX_WCHAR_SIZE;
        info.lpstrTitle = L"Load new model";
        info.nMaxFileTitle = sizeof(info.lpstrTitle);
        info.dwReserved = OFN_EXPLORER;
        info.Flags = OFN_NOCHANGEDIR;
        memcpy(g_wcCleanFileName, L"../../Media", sizeof(L"../../Media"));

        bResult = GetOpenFileName(&info);
    }

    if( bResult )
    {
        g_MeshScene.Destroy();
        // setup the camera view parameters
        static const D3D10_INPUT_ELEMENT_DESC mesh_layout[] =
        {
            { "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D10_INPUT_PER_VERTEX_DATA, 0 },
            { "NORMAL", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 12, D3D10_INPUT_PER_VERTEX_DATA, 0 },
            { "TEXCOORD", 0, DXGI_FORMAT_R32G32_FLOAT, 0, 24, D3D10_INPUT_PER_VERTEX_DATA, 0 },
        };
        if (g_MeshScene.Create(DXUTGetD3D10Device(), g_wcCleanFileName, (D3D10_INPUT_ELEMENT_DESC*)mesh_layout, 3) != S_OK)
        {
            MessageBox(DXUTGetHWND(), L"Could not load geometry from sunclock.x", L"Error", MB_OK);
            exit(0);
        }
        static bool bFirstTime = true;
        D3DXVECTOR3 vDiag = g_MeshScene.m_vBox[0] - g_MeshScene.m_vBox[1];
        if (bFirstTime || bNeedUI)
        {
            D3DXVECTOR3 vEye = bFirstTime ? D3DXVECTOR3(-57.089104, 33.029869, -58.556995) : (g_MeshScene.m_vBox[0] + vDiag);
            D3DXVECTOR3 vTmp2 = ((g_MeshScene.m_vBox[0] + g_MeshScene.m_vBox[1]) / 2);
            if (bFirstTime)
            {
                g_LCamera.SetModelRot(D3DXMATRIX(0.18668720, 0.60933846, -0.77061951, 0.00000000,
                                                 -0.44507933, -0.64683354, -0.61928248, 0.00000000,
                                                 -0.87581515, 0.45859894, 0.15044841, 0.00000000,
                                                 0.00000000, 0.00000000, 1.0000000, 1.0000008));
            }
            bFirstTime = false;
            g_Camera.SetViewParams(&vEye, &vTmp2);
        }
        g_Camera.SetModelCenter((g_MeshScene.m_vBox[0] + g_MeshScene.m_vBox[1]) / 2);
        g_LCamera.SetModelCenter((g_MeshScene.m_vBox[0] + g_MeshScene.m_vBox[1]) / 2);
        g_fSceneBoxDiag = sqrt(D3DXVec3Dot(&vDiag, &vDiag));
    }
}
static void InitApp()
{
    g_D3DSettingsDlg.Init( &g_DialogResourceManager );
    g_HUD.Init( &g_DialogResourceManager );
    g_SampleUI.Init( &g_DialogResourceManager );

    g_HUD.SetCallback( OnGUIEvent ); int iY = 10; 
    g_HUD.AddButton( IDC_TOGGLEFULLSCREEN, L"Toggle full screen", 35, iY, 125, 22 );
    g_HUD.AddButton( IDC_TOGGLEREF, L"Toggle REF (F3)", 35, iY += 24, 125, 22, VK_F3 );
    g_HUD.AddButton( IDC_CHANGEDEVICE, L"Change device (F2)", 35, iY += 24, 125, 22, VK_F2 );
    g_HUD.AddButton( IDC_LOADNEWMODEL, L"Load new model", 35, iY += 24, 125, 22 );

    g_SampleUI.EnableKeyboardInput( true );
    g_SampleUI.SetCallback( OnGUIEvent );
    iY = 10;
    g_SampleUI.AddStatic( IDC_SHADOW_ALGORITHM_LABEL, L"Shadow algorithm:", 35, iY, 125, 22 );
    CDXUTComboBox *pComboBox;
    g_SampleUI.AddComboBox( IDC_SHADOW_ALGORITHM, 35, iY += 20, 125, 30, 0, false, &pComboBox);
    pComboBox->AddItem(L"Accurate", NULL);
    pComboBox->AddItem(L"Fast", NULL);
    g_SampleUI.AddStatic( IDC_LIGHT_SIZE_LABEL, L"Light source size:", 35, iY += 25, 125, 22 );
    g_SampleUI.AddSlider( IDC_LIGHT_SIZE, 35, iY += 20, 124, 22, 0, 100, 0 );
    g_SampleUI.AddCheckBox( IDC_BTEXTURED, L"Enable Texturing", 35, iY += 25, 124, 22, true);
}
static void RenderText()
{
    CDXUTTextHelper txtHelper( g_pFont10, g_pSprite10, 15 );
    txtHelper.Begin();
    txtHelper.SetInsertionPos( 5, 5 );
    txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 0.0f, 1.0f ) );
    txtHelper.DrawTextLine( DXUTGetFrameStats(true) );
    txtHelper.DrawTextLine( DXUTGetDeviceStats() );
    txtHelper.DrawTextLine( L"Left mouse button to control light" );
    txtHelper.DrawTextLine( L"Right mouse button and mouse wheel to control camera" );
    txtHelper.DrawTextLine( L"F1 to toggle UI on/off" );
    txtHelper.End();
}

HRESULT CreatePipeBuffer(ID3D10Device* pDev10, UINT uiNumSpawnPoints, UINT uiMaxFaces);
HRESULT CreateRandomTexture(ID3D10Device* pDev10);
HRESULT GenerateTriCenterBuffer( ID3D10Device* pDev10, LPCTSTR szMesh, UINT *puNumFaces );
HRESULT LoadTextureArray( ID3D10Device* pDev10, LPCTSTR* szTextureNames, int iNumTextures, ID3D10Texture2D** ppTex2D, ID3D10ShaderResourceView** ppSRV);

//--------------------------------------------------------------------------------------
// Entry point to the program. Initializes everything and goes into a message processing 
// loop. Idle time is used to render the scene.
//--------------------------------------------------------------------------------------
int WINAPI wWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow )
{
    // Enable run-time memory check for debug builds.
#if defined(DEBUG) | defined(_DEBUG)
    _CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif

    // DXUT will create and use the best device (either D3D9 or D3D10) 
    // that is available on the system depending on which D3D callbacks are set below

    // Set DXUT callbacks
    DXUTSetCallbackDeviceChanging( ModifyDeviceSettings );
    DXUTSetCallbackMsgProc( MsgProc );
    DXUTSetCallbackKeyboard( OnKeyboard );
    DXUTSetCallbackFrameMove( OnFrameMove );
    DXUTSetCallbackD3D10DeviceAcceptable( IsD3D10DeviceAcceptable );
    DXUTSetCallbackD3D10DeviceCreated( OnD3D10CreateDevice );
    DXUTSetCallbackD3D10SwapChainResized( OnD3D10SwapChainResized );
    DXUTSetCallbackD3D10FrameRender( OnD3D10FrameRender );
    DXUTSetCallbackD3D10SwapChainReleasing( OnD3D10SwapChainReleasing );
    DXUTSetCallbackD3D10DeviceDestroyed( OnD3D10DestroyDevice );

    HRESULT hr;
    V_RETURN(DXUTSetMediaSearchPath(L"..\\Source\\SoftShadows"));

    InitApp();

    DXUTInit( true, true, NULL ); // Parse the command line, show msgboxes on error, no extra command line params
    DXUTSetCursorSettings( true, true ); // Show the cursor and clip it when in full screen
    DXUTCreateWindow( L"SoftShadows" );
    DXUTCreateDevice( true, 640, 480 );
    DXUTMainLoop(); // Enter into the DXUT render loop

    return DXUTGetExitCode();
}
//--------------------------------------------------------------------------------------
// Called right before creating a D3D9 or D3D10 device, allowing the app to modify the device settings as needed
//--------------------------------------------------------------------------------------
bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext )
{
    // For the first device created if its a REF device, optionally display a warning dialog box
    static bool s_bFirstTime = true;
	pDeviceSettings->d3d10.sd.SampleDesc.Count = 1;
	pDeviceSettings->d3d10.sd.SampleDesc.Quality = 0;
    if( s_bFirstTime )
    {
        s_bFirstTime = false;
        if((DXUT_D3D9_DEVICE == pDeviceSettings->ver && pDeviceSettings->d3d9.DeviceType == D3DDEVTYPE_REF) ||
            (DXUT_D3D10_DEVICE == pDeviceSettings->ver && pDeviceSettings->d3d10.DriverType == D3D10_DRIVER_TYPE_REFERENCE))
            DXUTDisplaySwitchingToREFWarning( pDeviceSettings->ver );
    }

    return true;
}
//--------------------------------------------------------------------------------------
// Handle updates to the scene.  This is called regardless of which D3D API is used
//--------------------------------------------------------------------------------------
void CALLBACK OnFrameMove( double fTime, float fElapsedTime, void* pUserContext )
{
    // update the camera's position based on user input 
    g_Camera.FrameMove(fElapsedTime);
    g_LCamera.FrameMove(fElapsedTime);
}
//--------------------------------------------------------------------------------------
// Handle messages to the application
//--------------------------------------------------------------------------------------
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing, void* pUserContext )
{
    // Pass messages to dialog resource manager calls so GUI state is updated correctly
    *pbNoFurtherProcessing = g_DialogResourceManager.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;

    // Pass messages to settings dialog if its active
    if( g_D3DSettingsDlg.IsActive() )
    {
        g_D3DSettingsDlg.MsgProc( hWnd, uMsg, wParam, lParam );
        return 0;
    }

    // Give the dialogs a chance to handle the message first
    *pbNoFurtherProcessing = g_HUD.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;
    *pbNoFurtherProcessing = g_SampleUI.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;

    static unsigned MouseMask = 0;
    if (uMsg == WM_LBUTTONDOWN || uMsg == WM_LBUTTONDBLCLK)
    { MouseMask |= 1; }
    if (uMsg == WM_LBUTTONUP)
    { MouseMask &= ~1; }

    // Pass all remaining windows messages to camera so it can respond to user input
    if (MouseMask & 1) // left button pressed
    { g_LCamera.HandleMessages( hWnd, uMsg, wParam, lParam ); }
    else g_Camera.HandleMessages( hWnd, uMsg, wParam, lParam );

    return 0;
}
//--------------------------------------------------------------------------------------
// Handle key presses
//--------------------------------------------------------------------------------------
void CALLBACK OnKeyboard(UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext)
{
    if( !bKeyDown )	return;
    switch( nChar )
    {
    case VK_F1:
        g_bShowUI = !g_bShowUI;
        break;
    }
}
//--------------------------------------------------------------------------------------
// Handles the GUI events
//--------------------------------------------------------------------------------------
void CALLBACK OnGUIEvent(UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext)
{
    switch( nControlID )
    {
    case IDC_TOGGLEFULLSCREEN: DXUTToggleFullScreen(); break;
    case IDC_TOGGLEREF: DXUTToggleREF(); break;
    case IDC_LOADNEWMODEL: LoadNewModel(true); break;
    case IDC_CHANGEDEVICE:
        g_D3DSettingsDlg.SetActive( !g_D3DSettingsDlg.IsActive() );
        break;
    case IDC_LIGHT_SIZE:
        ssmap.m_fFilterSize = (float)g_SampleUI.GetSlider(IDC_LIGHT_SIZE)->GetValue() / 100;
        break;
    case IDC_SHADOW_ALGORITHM:
        ssmap.bAccurateShadow = (((CDXUTComboBox*)pControl)->GetSelectedIndex() == 0);
        break;
    }    
}
//--------------------------------------------------------------------------------------
// Reject any D3D10 devices that aren't acceptable by returning false
//--------------------------------------------------------------------------------------
bool CALLBACK IsD3D10DeviceAcceptable(UINT Adapter, UINT Output, D3D10_DRIVER_TYPE DeviceType, DXGI_FORMAT BackBufferFormat, bool bWindowed, void* pUserContext)
{
    return true;
}
//--------------------------------------------------------------------------------------
// Create any D3D10 resources that aren't dependant on the back buffer
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnD3D10CreateDevice(ID3D10Device* pDev10, const DXGI_SURFACE_DESC *pBackBufferSurfaceDesc, void* pUserContext)
{
    HRESULT hr;

    V_RETURN(DXUTSetMediaSearchPath(L"..\\Source\\SoftShadows"));
    V_RETURN(g_DialogResourceManager.OnD3D10CreateDevice(pDev10));
    V_RETURN(g_D3DSettingsDlg.OnD3D10CreateDevice(pDev10));
    V_RETURN(D3DX10CreateFont(pDev10, 15, 0, FW_BOLD, 1, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, L"Arial", &g_pFont10));

    g_SampleUI.GetSlider(IDC_LIGHT_SIZE)->SetValue((int)(ssmap.m_fFilterSize * 100.0));
	g_SampleUI.GetComboBox(IDC_SHADOW_ALGORITHM)->SetSelectedByIndex(ssmap.bAccurateShadow == true ? 0 : 1);
    V_RETURN(D3DX10CreateSprite(pDev10, 512, &g_pSprite10));

    LoadNewModel();

    // setup the camera view parameters
    static const D3D10_INPUT_ELEMENT_DESC mesh_layout[] =
    {
        { "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D10_INPUT_PER_VERTEX_DATA, 0 },
        { "NORMAL", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 12, D3D10_INPUT_PER_VERTEX_DATA, 0 },
        { "TEXCOORD", 0, DXGI_FORMAT_R32G32_FLOAT, 0, 24, D3D10_INPUT_PER_VERTEX_DATA, 0 },
    };
    if (g_MeshLight.Create(pDev10, L"SoftShadows\\arrow.x", (D3D10_INPUT_ELEMENT_DESC*)mesh_layout, 3) != S_OK)
    {
        MessageBox(DXUTGetHWND(), L"Could not load geometry from arrow.x", L"Error", MB_OK);
        exit(0);
    }

    // create effect
    WCHAR str[MAX_PATH];
    V_RETURN(DXUTFindDXSDKMediaFileCch(str, MAX_PATH, L"softshadows.fx"));
    ID3D10Blob *pErrors;
    if (D3DX10CreateEffectFromFile(str, NULL, NULL, "fx_4_0", D3D10_SHADER_SKIP_OPTIMIZATION, 0, pDev10, NULL, NULL, &g_pEffect, &pErrors, &hr) != S_OK)
    {
        MessageBoxA(NULL, (char *)pErrors->GetBufferPointer(), "Compilation error", MB_OK);
        exit(0);
    }

    D3D10_PASS_DESC PassDesc;
    V_RETURN(g_pEffect->GetTechniqueByName("RenderAcc")->GetPassByIndex(0)->GetDesc(&PassDesc));
    V_RETURN(pDev10->CreateInputLayout(mesh_layout, 3, PassDesc.pIAInputSignature, PassDesc.IAInputSignatureSize, &g_pMaxLayout));

    D3DXVECTOR3 vTmp = D3DXVECTOR3(1, 2, 3);
    D3DXVec3Normalize(&g_vLightDir, &vTmp);

    SAFE_RELEASE(g_pRenderState);
    D3D10_RASTERIZER_DESC RasterizerState;
    RasterizerState.FillMode = D3D10_FILL_SOLID;
    RasterizerState.CullMode = D3D10_CULL_FRONT;
    RasterizerState.FrontCounterClockwise = true;
    RasterizerState.DepthBias = false;
    RasterizerState.DepthBiasClamp = 0;
    RasterizerState.SlopeScaledDepthBias = 0;
    RasterizerState.DepthClipEnable = true;
    RasterizerState.ScissorEnable = false;
    RasterizerState.MultisampleEnable = false;
    RasterizerState.AntialiasedLineEnable = false;
    V(pDev10->CreateRasterizerState(&RasterizerState, &g_pRenderState));

    return S_OK;
}
//--------------------------------------------------------------------------------------
// Create any D3D10 resources that depend on the back buffer
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnD3D10SwapChainResized( ID3D10Device* pDev10, IDXGISwapChain *pSwapChain, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
    HRESULT hr = S_OK;

    V_RETURN(g_DialogResourceManager.OnD3D10ResizedSwapChain(pDev10, pBackBufferSurfaceDesc));

    // setup the camera projection parameters
    g_Camera.SetWindow(pBackBufferSurfaceDesc->Width, pBackBufferSurfaceDesc->Height);
    g_LCamera.SetWindow(pBackBufferSurfaceDesc->Width, pBackBufferSurfaceDesc->Height);

    g_HUD.SetLocation(pBackBufferSurfaceDesc->Width-170, 0);
    g_HUD.SetSize(170, 170);

    g_SampleUI.SetLocation( pBackBufferSurfaceDesc->Width-170, pBackBufferSurfaceDesc->Height-300 );
    g_SampleUI.SetSize( 170, 300 );

    return hr;
}
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10FrameRender(ID3D10Device* pDev10, double fTime, float fElapsedTime, void* pUserContext)
{
    // compute view matrix
    D3DXMATRIX mTmp, mWorldView, mWorldViewProj, mWorldViewInv;
    D3DXMatrixInverse(&mTmp, NULL, g_Camera.GetWorldMatrix());
    D3DXMatrixMultiply(&mWorldView, &mTmp, g_Camera.GetViewMatrix());

    // correct near/far clip planes according to camera location
    const DXGI_SURFACE_DESC *pBackBufferSurfaceDesc = DXUTGetDXGIBackBufferSurfaceDesc();
    float fAspectRatio = pBackBufferSurfaceDesc->Width / (FLOAT)pBackBufferSurfaceDesc->Height;
    { // use slightly increased box to account for light mesh
        D3DXVECTOR3 vBox[2];
        vBox[0] = (g_MeshScene.m_vBox[1] - g_MeshScene.m_vBox[0]);
        vBox[1] = g_MeshScene.m_vBox[1] + vBox[0];
        vBox[0] = g_MeshScene.m_vBox[0] - vBox[0];
        g_Camera.SetProjParams(D3DX_PI/3, fAspectRatio, mWorldView, vBox);
    }

#if 0
    // if you want to fix camera position (for fps testing for example)
    mWorldView._11 = 0.43911904;
	mWorldView._12 = 0.89841902;
	mWorldView._13 = 0.0042110789;
	mWorldView._14 = 0.00000000;
    mWorldView._21 = 0.20170230;
    mWorldView._22 = -0.094016090;
	mWorldView._23 = -0.97492415;
	mWorldView._24 = 0.00000000;
	mWorldView._31 = -0.87549448;
	mWorldView._32 = 0.42895710;
	mWorldView._33 = -0.22249743;
	mWorldView._34 = 0.00000000;
	mWorldView._41 = -1.6029959;
	mWorldView._42 = 0.63748634;
	mWorldView._43 = 65.321449;
	mWorldView._44 = 1.0000000;
#endif
    D3DXMatrixMultiply(&mWorldViewProj, &mWorldView, g_Camera.GetProjMatrix());
    D3DXMatrixInverse(&mWorldViewInv, NULL, &mWorldView);

    // clear depth and color
    ID3D10DepthStencilView* pDSV = DXUTGetD3D10DepthStencilView();
    pDev10->ClearDepthStencilView( pDSV, D3D10_CLEAR_DEPTH, 1.0, 0);
    ID3D10RenderTargetView* pRTV = DXUTGetD3D10RenderTargetView();
    float ClearColor[4] = { 0, 0, 0.5, 1 };
    pDev10->ClearRenderTargetView(pRTV, ClearColor);

    if( g_D3DSettingsDlg.IsActive() )
    {
        g_D3DSettingsDlg.OnRender( fElapsedTime );
        return;
    }

    D3DXMATRIX mLightView;
    // here we compute light viewprojection so that light oversees the whole scene
    {
        D3DXMATRIX mTranslate;
        D3DXMatrixTranslation(&mTranslate, 0, 0, g_fSceneBoxDiag * 4);
        D3DXMatrixMultiply(&mLightView, g_LCamera.GetWorldMatrix(), &mTranslate);
        // compute projection matrix
        D3DXVECTOR3 BBox[2];
        BBox[0][0] = BBox[0][1] = BBox[0][2] =  1E38;
        BBox[1][0] = BBox[1][1] = BBox[1][2] = -1E38;
        for (int i = 0; i < 8; ++i)
        {
            D3DXVECTOR3 v, v1;
            v[0] = g_MeshScene.m_vBox[(i & 1) ? 0 : 1][0];
            v[1] = g_MeshScene.m_vBox[(i & 2) ? 0 : 1][1];
            v[2] = g_MeshScene.m_vBox[(i & 4) ? 0 : 1][2];
            D3DXVec3TransformCoord(&v1, &v, &mLightView);
            for (int j = 0; j < 3; ++j)
            {
                BBox[0][j] = min(BBox[0][j], v1[j]);
                BBox[1][j] = max(BBox[1][j], v1[j]);
            }
        }
        // expand the box a bit
        for (int j = 0; j < 3; ++j)
        {
            BBox[0][j] -= (FLOAT)(abs(BBox[0][j]) * 1e-3);
            BBox[1][j] += (FLOAT)(abs(BBox[1][j]) * 1e-3);
        }
        D3DXMatrixPerspectiveLH(&ssmap.mLightProj, (FLOAT)max(abs(BBox[0][0]), (FLOAT)abs(BBox[1][0])) * 2, (FLOAT)max(abs(BBox[0][1]), (FLOAT)abs(BBox[1][1])) * 2, (FLOAT)max(0.0001, BBox[0][2]), (FLOAT)max(0.0002, BBox[1][2]));
        D3DXMatrixMultiply(&ssmap.mLightViewProj, &mLightView, &ssmap.mLightProj);
        HRESULT hr;
        V(g_pEffect->GetVariableByName("mViewProj")->AsMatrix()->SetMatrix((float *)&ssmap.mLightViewProj));
        V(g_pEffect->GetVariableByName("mLightView")->AsMatrix()->SetMatrix((float *)&mLightView));
    }

    // render shadow map
    ssmap.Render(pDev10, &g_MeshScene, g_pEffect);

    HRESULT hr;
    V(g_pEffect->GetVariableByName("mViewProj")->AsMatrix()->SetMatrix((float *)&mWorldViewProj));
    {
        D3DXVECTOR3 vLgtPos, vZero = D3DXVECTOR3(0, 0, 0);
        D3DXMATRIX mLightViewInv;
        D3DXMatrixInverse(&mLightViewInv, NULL, &mLightView);
        D3DXVec3TransformCoord(&vLgtPos, &vZero, &mLightViewInv);
        V(g_pEffect->GetVariableByName("g_vLightPos")->AsVector()->SetRawValue(&vLgtPos, 0, sizeof(vLgtPos)));
    }
    {
        unsigned iTmp = g_SampleUI.GetCheckBox(IDC_BTEXTURED)->GetChecked();
        V(g_pEffect->GetVariableByName("bTextured")->AsScalar()->SetRawValue(&iTmp, 0, sizeof(iTmp)));
        D3DXVECTOR4 vTmp = D3DXVECTOR4(1, 1, (float)iTmp, 1);
        V(g_pEffect->GetVariableByName("g_vLightFlux")->AsVector()->SetRawValue(&vTmp, 0, sizeof(D3DXVECTOR4)));
        V(g_pEffect->GetVariableByName("g_vMaterialKd")->AsVector()->SetRawValue(&vTmp, 0, sizeof(D3DXVECTOR4)));
    }

    pDev10->RSSetState(g_pRenderState);

    // do z-only pass and fill z-buffer
    pDev10->IASetInputLayout(ssmap.m_pDepthLayout);
    g_MeshScene.Render(pDev10, ssmap.m_pDRenderTechnique);

    // render with shader
    pDev10->IASetInputLayout(g_pMaxLayout);
    ID3D10EffectShaderResourceVariable *pTexture = g_pEffect->GetVariableByName("DiffuseTex")->AsShaderResource();
    if (ssmap.bAccurateShadow)
    {
        g_MeshScene.Render(pDev10, g_pEffect->GetTechniqueByName("RenderAcc"), pTexture);
    }
    else
    {
        g_MeshScene.Render(pDev10, g_pEffect->GetTechniqueByName("RenderFast"), pTexture);
    }

    // render light
    {
        D3DXMATRIX mLightViewInv;
        D3DXMatrixInverse(&mLightViewInv, NULL, &mLightView);
        D3DXMATRIX mLightViewInvWorldViewProj;
        D3DXMatrixMultiply(&mLightViewInvWorldViewProj, &mLightViewInv, &mWorldViewProj);
        V(g_pEffect->GetVariableByName("mViewProj")->AsMatrix()->SetMatrix((float *)&mLightViewInvWorldViewProj));
        g_pEffect->GetTechniqueByName("RenderNoShadows")->GetPassByIndex(0)->Apply(0);
        pDev10->RSSetState(g_pRenderState);
        pDev10->IASetInputLayout(g_pMaxLayout);
        g_MeshLight.Render(pDev10);
    }

    // render UI
    if (g_bShowUI)
    {
        RenderText();
        g_SampleUI.OnRender(fElapsedTime);
        g_HUD.OnRender(fElapsedTime);
    }

#if 0
  static int g_Frame = 0;
  IDXGISwapChain* pSwapChain = DXUTGetDXGISwapChain();
  ID3D10Texture2D* pRT;
  pSwapChain->GetBuffer(0, __uuidof(pRT), reinterpret_cast<void**>(&pRT));
  WCHAR filename[32];
  StringCchPrintf(filename, 100, L"d:\\screenshot%.3d.bmp", g_Frame); 
  D3DX10SaveTextureToFile(pRT, D3DX10_IFF_BMP, filename);
  pRT->Release();
  ++g_Frame;
#endif
}
//--------------------------------------------------------------------------------------
// Release D3D10 resources created in OnD3D10ResizedSwapChain 
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10SwapChainReleasing( void* pUserContext )
{
    g_DialogResourceManager.OnD3D10ReleasingSwapChain();
    ssmap.OnWindowResize();
}
//--------------------------------------------------------------------------------------
// Release D3D10 resources created in OnD3D10CreateDevice 
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10DestroyDevice( void* pUserContext )
{
    ssmap.OnDestroy();

    g_DialogResourceManager.OnD3D10DestroyDevice();
    g_D3DSettingsDlg.OnD3D10DestroyDevice();

    SAFE_RELEASE(g_pFont10);
    SAFE_RELEASE(g_pSprite10);
    SAFE_RELEASE(g_pEffect);
    SAFE_RELEASE(g_pRenderState);
    SAFE_RELEASE(g_pMaxLayout);

    g_MeshScene.Destroy();
    g_MeshLight.Destroy();
}
